require 'yaml'
require 'log4r'
include Log4r
require 'fcgi'

require 'iowa/Association'
require 'iowa/Context'
require 'iowa/Element'
require 'iowa/Tag'
require 'iowa/DynamicElements'
require 'iowa/Form'
require 'iowa/Component'
require 'iowa/ComponentProxy'
require 'iowa/TemplateParser'
require 'iowa/BindingsParser'
require 'iowa/KeyValueCoding'
require 'iowa/PageStore'
require 'iowa/Session'
require 'iowa/Application'
require 'iowa/config'
require 'iowa/Request'
require 'iowa/Email'

require 'socket'

def unescape(str)
	str.gsub(/\+/, ' ').gsub(/%([0-9a-fA-F]{2})/){ [$1.hex].pack("c") }
end

class Hash
	# A simple little routine to downcase any String used as a key
	# in a hash or any hash nested as value somewhere within the hash.

	def downcase_keys
		temp = self.clone
		temp.each do |key,value|
			if key.class.to_s == "String"
				temp.delete key
				if value.respond_to? :downcase_keys
					temp[key.downcase] = value.downcase_keys 
				else
					temp[key.downcase] = value
				end
			end
		end

		temp
	end
end

module Iowa

	@@app = nil

	# An Iowa appplication's configuration is given in a plain text file
	# formatted as YAML (http://www.yaml.org).

	def Iowa.readConfiguration(configuration_file)
		begin
			cf = File.open configuration_file

			rescue Exception => exception
				puts "Failure while accessing the configuration file (#{configuration_file})\n"
				raise
		end

		@@config = YAML::load(cf).downcase_keys

		checkConfiguration
	end


	# Forks the process.  The parent outputs the PID of the child (so that
	# an external invocation process can capture this information) and then
	# exits.  The child sets inself as it's session leader and returns.

	def Iowa.startDaemon(configuration_file)
		readConfiguration(configuration_file)
		startLogging

	 if (child_pid = fork)
	   puts "PID #{child_pid}"
	   exit
	 end
		mylog = Logger['iowa_log']
	 Process.setsid
		mylog.info "Process forked and daemonized as PID #{Process.pid}."
	end

	# Check that the minimum required information is present in the
	# configuration, and set the defaults for information that was
	# omitted. Exceptions will be raised if socket information is
	# missing, or if either socket or logging information seems to
	# have errors.

	def Iowa.checkConfiguration
		# Make sure socket info was provided
		raise "A socket specification must be provided in the configuration file." unless @@config.has_key? 'socket'
		raise "Iowa supports both TCP and Unix domain sockets.  Please specify the socket to use within the 'socket' section of the configuration file." unless (@@config['socket'].class.to_s == 'Hash' and (@@config['socket'].has_key? 'path' or @@config['socket'].has_key? 'port'))
		raise "Please define only one of a TCP or a Unix domain socket in the configuration file." if @@config['socket'].has_key? 'path' and @@config['socket'].has_key? 'port'

		# Verify that the socket information is valid.  i.e. that the provided
		# port is a number, that the hostname for a TCP socket doesn't have
		# funny characters in it, or that the path for a Unix socket is writeable.
		raise "A port number must be specified as an integer." unless @@config['socket'].has_key? 'port' and @@config['socket']['port'].class.to_s == 'Fixnum'
		raise "The socket hostname contains illegal characters." unless @@config['socket'].has_key? 'hostname' and @@config['socket']['hostname'] =~ /^[a-zA-Z0-9\.\-]*$/
		if @@config['socket'].has_key? 'path'
			begin
				sf = File.open(@@config['socket']['path'],'w')
				@@config['socket']['basename'] = sf.basename
				sf.close

			rescue Exception => exception
				puts "The path for a Unix socket must be writeable."
				raise
			end
		end

		# If any logging information has been specified, that needs to be
		# checked, as well.

		if @@config.has_key? 'logging'
			if @@config['logging'].has_key? 'basedir'
				raise "The logging basedir (#{@@config['logging']['basedir']}) must be a directory." unless FileTest.directory? @@config['logging']['basedir']
				raise "The logging basedir #{@@config['logging']['basedir']}) must be writeable." unless FileTest.writable? @@config['logging']['basedir']
			end
		
			raise "The minimum logging level (minlevel) must be an integer between 0 to 7, inclusively." if @@config['logging'].has_key? 'minlevel' and (@@config['logging']['minlevel'] < 0 or @@config['logging']['minlevel'] > 7)

			raise "The maximum log file size (maxsize) must be greater than zero." if @@config.has_key? 'maxsize' and @@config['maxsize'] <= 0

			raise "The maximum log file age (maxage) must be greater than zero." if @@config.has_key? 'maxage' and @@config['maxage'] <= 0
		end

		# Now apply any defaults for optional items which were not specified.

		@@config['logging'] = Hash.new unless @@config.has_key? 'logging'
		@@config['logging']['basedir'] = '/tmp' unless @@config['logging'].has_key? 'basedir'
		@@config['logging']['minlevel'] = 'info' unless @@config['logging'].has_key? 'minlevel'
		@@config['logging']['maxsize'] = 1024 * 1024 unless @@config['logging'].has_key? 'maxsize'
		@@config['logging']['maxage'] = 86400 unless @@config['logging'].has_key? 'maxage'
		if @@config['socket'].has_key? 'path'
			@@config['logging']['filename'] = "#{@@config['logging']['basedir']}/#{@@config['socket']['basename']}.log"
		else
			@@config['logging']['filename'] = "#{@@config['logging']['basedir']}/#{@@config['socket']['hostname']}:#{@@config['socket']['port']}.log"
		end
	end

	# Initialize logging mechanisms

	def Iowa.startLogging
		lcfg = @@config['logging']
		mylog = Logger.new 'iowa_log'
		myformat = PatternFormatter.new :pattern => '[%l] %C @ %d :: %M'
		mybaselog = RollingFileOutputter.new 'baselog', :maxsize => lcfg['maxsize'], :maxtime => lcfg['maxage'], :level => lcfg['minlevel'], :filename => lcfg['filename'], :trunc => false, :formatter => myformat
		mylog.outputters = mybaselog

		myApacheLikeLog = Logger.new 'iowa_apachelike_log'
		myApacheLikeFormat = PatternFormatter.new :pattern => '%M'
		mybaseApacheLikeLog = RollingFileOutputter.new 'apachelike_baselog', :maxsize => 999999999, :maxtime => 86400, :level => lcfg['minlevel'], :filename => lcfg['filename'] + '.apachelike', :trunc => false, :formatter => myApacheLikeFormat
		myApacheLikeLog.outputters = mybaseApacheLikeLog

		mylog.info 'Logging subsystem initialized.'
	end

	# Handle the communications coming in on the monitored socket,
	# create a context object from the data received, and then pass
	# the context information into the Application object for final
	# handling.  Exception handling is simply via capturing the Exception
	# and outputting a stack backtrace (this could be improved).

	def Iowa.handleConnection(fcgi_request)
		start_time = Time.now
		mylog = Logger['iowa_log']
		# Having two logfiles with much the same data in them is stupid.
		myapachelikelog = Logger['iowa_apachelike_log']

		request = Iowa::Request.new

		buffer = ''
		context = nil
		exception = nil
		exception = catch(:session_error) do
			begin
				context = Iowa::Context.new(request, buffer)
				app.handleRequest(context)
			rescue Exception => exception
			end
		end

		if exception.to_s != '' and exception.kind_of? Exception
			buffer << "<p>#{exception}<br/>"
			buffer << exception.backtrace.join("<br/>\n").to_s
			buffer << "</p>"
			mylog.warn "Execution Error: #{exception} :: " + exception.backtrace.join(".....").to_s
			status = 'EOUT'
		end

		begin
			request.content = buffer.to_s
			request.headers_out.each_pair {|k,v| fcgi_request.out.print "#{k}: #{v}"}
			fcgi_request.out.print request.content
			end_time = Time.now
		rescue Exception => exception
			mylog.error "Failure While Writing Outbound Data: #{exception}"
		end

		vhost = nil
		request.unparsed_uri =~ /http:\/\/([^\/]*)\//
		vhost = $1
		referer = request.headers_in['Referer'] != '' ? request.headers_in['Referer'] : request.uri
		myapachelikelog.info "#{request.remote_host} #{request.hostname} - [#{Time.now.strftime('%d/%b/%Y:%H:%M:%S')}] #{buffer.length} \"#{request.request_method} #{request.unparsed_uri} HTTP/1.1\" 200 #{buffer.length} \"#{referer}\" \"#{request.headers_in['User-Agent']}\" \"#{request.headers_in['Cookie']}\""
		mylog.info "#{start_time} (#{read_time.to_f - start_time.to_f}/#{end_time.to_f - start_time.to_f}) :: #{request.uri} #{status} #{buffer.length}B"
	end

	# Outputs the location of the socket being monitored, then enters the
	# event loop to wait for and handle connections.

	def Iowa.run()
		mylog = Logger['iowa_log']
		if @@config['socket'].has_key? 'port'
			socket_host = @@config['socket']['hostname']
			socket_port = @@config['socket']['port']
			begin
				# This should be configurable and runtime toggleable.
				TCPServer.do_not_reverse_lookup = true
				if socket_host.to_s != ''
					server = TCPserver.new(socket_port)
					mylog.info "Listening at localhost:#{socket_port}."
				else
					server = TCPserver.new(socket_host,socket_port)
					mylog.info "Listening at #{socket_host}:#{socket_port}."
				end
			rescue Exception => exception
				mylog.fatal "Failure while attempting to establish a TCP socket at #{socket_host}:#{socket_port} : #{exception}"
				raise
			end

		else
			begin
				server = UNIXServer.new(@@config['socket']['path'])
				File.chmod(0666, @@config['socket']['path'])  #This seems stupid.
			rescue Exception => exception
				mylog.fatal "Failure while attemting to establish a Unix socket at #{@@config['socket']['path']} : #{exception}"
				raise
			end
		end

		# initialLoad performs the first loading of the application files.
		# This has been seperated from reloadModified in order to
		# potentially provide a hook that can be used to improve
		# overall loading performance.  If we just could persistently
		# cache a parsed template....
		app.initialLoad()

		begin
			mylog.info 'Entering the main event loop...'
			FCGI.each do |request|
				handleConnection request
				request.finish
			end
		rescue Exception => exception
			mylog.fatal "Catastrophic failure in main event loop: #{exception} :: " + exception.backtrace.join(".....").to_s
		ensure
			server.close
			File.delete(@@config['socket']['path']) if @@config['socket'].has_key? 'path'
		end
	end

	def Iowa.app
		mylog = Logger['iowa_log']
		if @@app
			@@app
		else
			@@app = Application.newApplication(Dir.getwd)
		end
	end
end
